/***************************************************************************
                       DSLib v2.0.0 P-I regulator example
                             -------------------
    copyright            : (C) 2003 by Ivan Martinez
    email                : imr@oersted.dtu.dk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2.1 of the License, or   *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <dynamicsystem.h>
#include <blockvariable.h>
#include <function.h>
#include <periodicscheduler.h>
#include <fstream>
#include <comediinput.h>
#include <comedioutput.h>
#include <comedi.h>
#include <rtaisharedmemorywriter.h>
#include <rtlfiforeader.h>

// One second is 1e9 nanoseconds (time unit required by functions)
#define SECONDS 1e9
// Variables are logged in one every 1/LOGGING_SAMPLING_RATE executions
// of the scheduler
#define LOGGING_SAMPLING_RATE 1
// Realtime fifos will be able to store FIFOS_SAMPLE samples of their
// variable
#define FIFOS_LENGTH 1000
// Period of the execution of the scheduler, in seconds
#define EXECUTION_PERIOD	0.01

int main(int argc, char *argv[])
{
	// Defining a dynamic system
	CDynamicSystem Regulator;
	Regulator.SetTime((long long int) 10 * SECONDS);
	Regulator.SetRTClockResolution((long long int) 0.1 * SECONDS);
	Regulator.SetMakefileInstallCommands(
		"insmod comedi\n"
		"insmod kcomedilib\n"
		"insmod comedi_fc\n"
		"insmod comedi_test\n"
		"comedi_config /dev/comedi0 comedi_test 5,1\n"
		"comedi_config /dev/comedi3 comedi_test 5,1\n");
	Regulator.SetMakefileRemoveCommands(
		"-rmmod comedi_test\n"
		"-rmmod comedi_fc\n"
		"-rmmod kcomedilib\n"
		"-rmmod comedi\n");

	CPeriodicScheduler* pScheduler = new CPeriodicScheduler();
	Regulator.AddComponent(pScheduler);
	pScheduler->SetIdentifier("L");
	pScheduler->SetPeriod(0.1 * SECONDS);
	pScheduler->SetPriority(0);

	CRTLFifoReader* pRTLFRScheduler = new CRTLFifoReader();
	pScheduler->AddComponent(pRTLFRScheduler);
	pRTLFRScheduler->SetIdentifier("logging_fifo");
	pRTLFRScheduler->SetNumber(3);
	pRTLFRScheduler->SetSamplingRate(LOGGING_SAMPLING_RATE);
	pRTLFRScheduler->SetLength(FIFOS_LENGTH);

	CRTAISharedMemoryWriter* pRTAISMWScheduler = new CRTAISharedMemoryWriter();
	pScheduler->AddComponent(pRTAISMWScheduler);
	pRTAISMWScheduler->SetIdentifier("shared_memory");
	pRTAISMWScheduler->SetMemoryIdentifier(1);

	// Comedi input
	CComediInput * pComediInput = new CComediInput();
	pScheduler->AddComponent(pComediInput);
	pComediInput->SetIdentifier("comedi_input");
	pComediInput->SetDeviceFile("/dev/comedi0");
	pComediInput->SetSubdevice(0);
	pComediInput->SetChannel(0);
	pComediInput->SetRange(0);
	pComediInput->SetReference(AREF_GROUND);
	// Following two lines set the range of the block's output from -10 to +10
	pComediInput->SetOutputRange(20);
	pComediInput->SetOutputOffset(-10);
	pComediInput->Output()->SetRead(true);

	// Summing Point
	CFunction* pFSummingPoint = new CFunction();
	pScheduler->AddComponent(pFSummingPoint);
	pFSummingPoint->SetIdentifier("summing_point");
	pFSummingPoint->SetRunCode(
		"error = reference - system_output;\n");
	pFSummingPoint->AddBlockVariable("reference", "float");
	pFSummingPoint->AddInput("system_output", "float");
	pFSummingPoint->Input("system_output")->Connect(
		pComediInput->Output());
	pFSummingPoint->AddOutput("error", "float");
	pFSummingPoint->BlockVariable("reference")->SetRead(true);
	pFSummingPoint->BlockVariable("reference")->SetWrite(true);
	pFSummingPoint->Output("error")->SetRead(true);

	// Integrator
	CFunction* pIntegrator = new CFunction();
	pScheduler->AddComponent(pIntegrator);
	pIntegrator->SetIdentifier("Integrator");
	pIntegrator->SetRunCode(
		"state = 1 * state + 0.1 * input;\n"
		"output = state;\n");
	pIntegrator->AddInput("input", "float");
	pIntegrator->AddOutput("output", "float");
	pIntegrator->AddBlockVariable("state", "float");
	pIntegrator->Input("input")->Connect(
		pFSummingPoint->Output("error"));
	pIntegrator->Output("output")->SetRead(true);

	// Comedi Output
	CComediOutput * pComediOutput = new CComediOutput();
	pScheduler->AddComponent(pComediOutput);
	pComediOutput->SetIdentifier("ComediOutput");
	pComediOutput->SetDeviceFile("/dev/comedi3");
	pComediOutput->SetSubdevice(0);
	pComediOutput->SetChannel(0);
	pComediOutput->SetRange(0);
	pComediOutput->SetReference(AREF_GROUND);
	// Following two lines set the range of the block's input from -20 to +20
	pComediOutput->SetInputRange(20);
	pComediOutput->SetInputOffset(-10);
	pComediOutput->Input()->Connect(pIntegrator->Output("output"));

	// Once the definition of the dynamic system is complete, ArrangeAll must
	// be called.
	bool bNoError = Regulator.ArrangeAll();

	// If any error has been detected in the definition...
	if (!bNoError)
	{
		CMessages Messages;
		Regulator.AllReportMessages(&Messages);

		CMessages::m_TMessagesVector::const_iterator MessagesIterator;

		// The information about the detected errors is printed.
		for (MessagesIterator = Messages.Vector().begin();
			MessagesIterator != Messages.Vector().end();
			MessagesIterator++)
		{
			std::cout << "Error Number: " << (*MessagesIterator).Number() <<
				std::endl;
			std::cout << "Class with the error: " <<
				(*MessagesIterator).Location()->Class() << std::endl;
			std::cout << "Identifier of the class with the error: " <<

				(*MessagesIterator).Location()->Identifier() << std::endl;
			if ((*MessagesIterator).Location()->Container() != NULL)
				std::cout << "Identifier of the container of the class with the "
					"error: " <<
					(*MessagesIterator).Location()->Container()->Identifier() <<
						std::endl;
		}
	}
	// If there are no errors
	else
	{
		std::ofstream FileStream;
		// Generate read/write data structure
		FileStream.open("../../rtsystem/rw_types.h");
		Regulator.GenerateCode("./dslib-2/templates/rw_types/", FileStream);
		FileStream.close();
		std::cout << "rw_types.h generated" << std::endl;
		// Generate RTAI module
		FileStream.open("../../rtsystem/rtmodule.c");
		Regulator.GenerateCode("./dslib-2/templates/rtai_module/", FileStream);
		FileStream.close();
		std::cout << "rtmodule.c generated" << std::endl;
		// Generate C monitor
		FileStream.open("../../rtsystem/monitor.c");
		Regulator.GenerateCode("./dslib-2/templates/c_command_monitor/",
			FileStream);
		FileStream.close();
		std::cout << "monitor.c generated" << std::endl;
		// Generate Tcl/Tk intenface
		FileStream.open("../../rtsystem/monitor.tcl");
		Regulator.GenerateCode("./dslib-2/templates/tcltk_monitor_gui/",
			FileStream);
		FileStream.close();
		std::cout << "rtmodule.tcl generated" << std::endl;
		// Genetare RTAI Makefile
		FileStream.open("../../rtsystem/Makefile");
		Regulator.GenerateCode("./dslib-2/templates/rtai_makefile/", FileStream);
		FileStream.close();
		std::cout << "Makefile generated" << std::endl;
	}

	// The following deletes all the objects contained by Regulator,
	// directly or indirectly, so all the allocated memory is freed.
	Regulator.DeleteObjects();

	return EXIT_SUCCESS;
}
